#pragma once

#include <stdint.h>

// ft2_bmp_fonts.c
extern const uint8_t buttonGfxBMP[416];
extern const uint8_t font1BMP[3272];
extern const uint8_t font2BMP[10060];
extern const uint8_t font3BMP[584];
extern const uint8_t font4BMP[13604];
extern const uint8_t font6BMP[532];
extern const uint8_t font7BMP[472];
extern const uint8_t font8BMP[368]; // small font for piano key and Nibbles (snake food number)

// ft2_bmp_logo.c
extern const uint8_t ft2AboutLogoBMP[29204];
extern const uint8_t ft2ByBadgesBMP[1916];
extern const uint8_t ft2LogoBadgesBMP[6260];

// ft2_bmp_mouse.c
extern const uint8_t mouseCursorsBMP[3444];
extern const uint8_t mouseCursorBusyClockBMP[1728];
extern const uint8_t mouseCursorBusyGlassBMP[6136];

// ft2_bmp_nibbles.c
extern const uint8_t nibblesLogoBMP[2252];
extern const uint8_t nibblesStagesBMP[11088];

// ft2_bmp_instr.c
extern const uint8_t vibratoWaveformsBMP[468];
extern const uint8_t whitePianoKeysBMP[2580];
extern const uint8_t blackPianoKeysBMP[480];

// ft2_bmp_looppins.c
extern const uint8_t loopPinsBMP[6256];

// ft2_bmp_scopes.c
extern const uint8_t scopeRecBMP[120];
extern const uint8_t scopeMuteBMP[5068];

// ft2_bmp_midi.c
extern const uint8_t midiLogoBMP[1544];

// ft2_bmp_gui.c
extern const uint8_t checkboxGfxBMP[776];
extern const uint8_t radiobuttonGfxBMP[404];
