#pragma once

void drawTrimScreen(void);
void hideTrimScreen(void);
void showTrimScreen(void);
void toggleTrimScreen(void);
void setInitialTrimFlags(void);
void trimThreadDone(void);
void resetTrimSizes(void);
void cbTrimUnusedPatt(void);
void cbTrimUnusedInst(void);
void cbTrimUnusedSamp(void);
void cbTrimUnusedChans(void);
void cbTrimUnusedSmpData(void);
void cbTrimSmpsTo8Bit(void);
void pbTrimCalc(void);
void pbTrimDoTrim(void);
